@@ -10,7 +10,7 @@ Convert::BER::XS - I<very> low level BER en-/decoding
     or die "unable to decode SNMP message";
 
  # The above results in a data structure consisting of
- # (class, tag, flags, data)
+ #    (class, tag, flags, data)
  # tuples. Below is such a message, SNMPv1 trap
  # with a Cisco mac change notification.
  # Did you know that Cisco is in the news almost
@@ -38,6 +38,9 @@ Convert::BER::XS - I<very> low level BER en-/decoding
                      ]
                   ],
                   ...
+ # let's dump it, for debugging
+
+ ber_dump $ber, $Convert::BER::XS::SNMP_PROFILE;
 
  # let's decode it a bit with some helper functions
 
@@ -115,8 +118,8 @@ the right):
 ASN tag values (some of which are aliases, such as C<ASN_OID>). Their
 numerical value corresponds exactly to the numbers used in BER/X.690.
 
-      ASN_BOOLEAN ASN_INTEGER ASN_BIT_STRING ASN_OCTET_STRING ASN_NULL ASN_OBJECT_IDENTIFIER
-      ASN_OBJECT_DESCRIPTOR ASN_OID ASN_EXTERNAL ASN_REAL ASN_SEQUENCE ASN_ENUMERATED
+      ASN_BOOLEAN ASN_INTEGER ASN_BIT_STRING ASN_OCTET_STRING ASN_NULL ASN_OID
+      ASN_OBJECT_IDENTIFIER ASN_OBJECT_DESCRIPTOR ASN_EXTERNAL ASN_REAL ASN_SEQUENCE ASN_ENUMERATED
       ASN_EMBEDDED_PDV ASN_UTF8_STRING ASN_RELATIVE_OID ASN_SET ASN_NUMERIC_STRING
       ASN_PRINTABLE_STRING ASN_TELETEX_STRING ASN_T61_STRING ASN_VIDEOTEX_STRING ASN_IA5_STRING
       ASN_ASCII_STRING ASN_UTC_TIME ASN_GENERALIZED_TIME ASN_GRAPHIC_STRING ASN_VISIBLE_STRING
@@ -145,6 +148,7 @@ C<ber_decode> and the match helper functions:
 
       ber_decode ber-decode_prefix
       ber_is ber_is_seq ber_is_int ber_is_oid
+      ber_dump
 
 =item C<:encode>
 
@@ -410,7 +414,7 @@ use Exporter qw(import);
 our $VERSION;
 
 BEGIN {
-   $VERSION = '1.0';
+   $VERSION = 1.1;
    XSLoader::load __PACKAGE__, $VERSION;
 }
 
@@ -418,15 +422,16 @@ our %EXPORT_TAGS = (
    const_index => [qw(
       BER_CLASS BER_TAG BER_FLAGS BER_DATA
    )],
-   const_asn => [qw(
-      ASN_BOOLEAN ASN_INTEGER ASN_BIT_STRING ASN_OCTET_STRING ASN_NULL ASN_OBJECT_IDENTIFIER
-      ASN_OBJECT_DESCRIPTOR ASN_OID ASN_EXTERNAL ASN_REAL ASN_SEQUENCE ASN_ENUMERATED
+   const_asn_class => [qw(
+      ASN_UNIVERSAL ASN_APPLICATION ASN_CONTEXT ASN_PRIVATE
+   )],
+   const_asn_tag => [qw(
+      ASN_BOOLEAN ASN_INTEGER ASN_BIT_STRING ASN_OCTET_STRING ASN_NULL ASN_OID ASN_OBJECT_IDENTIFIER
+      ASN_OBJECT_DESCRIPTOR ASN_EXTERNAL ASN_REAL ASN_SEQUENCE ASN_ENUMERATED
       ASN_EMBEDDED_PDV ASN_UTF8_STRING ASN_RELATIVE_OID ASN_SET ASN_NUMERIC_STRING
       ASN_PRINTABLE_STRING ASN_TELETEX_STRING ASN_T61_STRING ASN_VIDEOTEX_STRING ASN_IA5_STRING
       ASN_ASCII_STRING ASN_UTC_TIME ASN_GENERALIZED_TIME ASN_GRAPHIC_STRING ASN_VISIBLE_STRING
       ASN_ISO646_STRING ASN_GENERAL_STRING ASN_UNIVERSAL_STRING ASN_CHARACTER_STRING ASN_BMP_STRING
-
-      ASN_UNIVERSAL ASN_APPLICATION ASN_CONTEXT ASN_PRIVATE
    )],
    const_ber_type => [qw(
       BER_TYPE_BYTES BER_TYPE_UTF8 BER_TYPE_UCS2 BER_TYPE_UCS4 BER_TYPE_INT
@@ -440,6 +445,7 @@ our %EXPORT_TAGS = (
    decode => [qw(
       ber_decode ber_decode_prefix
       ber_is ber_is_seq ber_is_int ber_is_oid
+      ber_dump
    )],
    encode => [qw(
       ber_encode
@@ -449,8 +455,124 @@ our %EXPORT_TAGS = (
 
 our @EXPORT_OK = map @$_, values %EXPORT_TAGS;
 
-$EXPORT_TAGS{all} = \@EXPORT_OK;
-$EXPORT_TAGS{const} = [map @{ $EXPORT_TAGS{$_} }, qw(const_index const_asn)];
+$EXPORT_TAGS{all}       = \@EXPORT_OK;
+$EXPORT_TAGS{const_asn} = [map @{ $EXPORT_TAGS{$_} }, qw(const_asn_class const_asn_tag)];
+$EXPORT_TAGS{const}     = [map @{ $EXPORT_TAGS{$_} }, qw(const_index const_asn)];
+
+our $DEFAULT_PROFILE = new Convert::BER::XS::Profile;
+
+$DEFAULT_PROFILE->_set_default;
+
+# additional SNMP application types
+our $SNMP_PROFILE = new Convert::BER::XS::Profile;
+
+$SNMP_PROFILE->set (ASN_APPLICATION, SNMP_IPADDRESS , BER_TYPE_IPADDRESS);
+$SNMP_PROFILE->set (ASN_APPLICATION, SNMP_COUNTER32 , BER_TYPE_INT);
+$SNMP_PROFILE->set (ASN_APPLICATION, SNMP_UNSIGNED32, BER_TYPE_INT);
+$SNMP_PROFILE->set (ASN_APPLICATION, SNMP_TIMETICKS , BER_TYPE_INT);
+$SNMP_PROFILE->set (ASN_APPLICATION, SNMP_OPAQUE    , BER_TYPE_IPADDRESS);
+
+=head2 DEBUGGING
+
+To aid debugging, you cna call the C<ber_dump> function to print a "nice"
+representation to STDOUT.
+
+=over
+
+=item ber_dump $tuple[, $profile[, $prefix]]
+
+In addition to specifying the BER C<$tuple> to dump, youc an also specify
+a C<$profile> and a C<$prefix> string that is printed in front of each line.
+
+If C<$profile> is C<$Convert::BER::XS::SNMP_PROFILE>, then C<ber_dump>
+will try to improve its output for SNMP data.
+
+The output usually contains three columns, the "human readable" tag, the
+BER type used to decode it, and the data value.
+
+This function is somewhat slow and uses a number of heuristics and tricks,
+so it really is only suitable for debug prints.
+
+Example output:
+
+   SEQUENCE
+   | OCTET_STRING     bytes  800063784300454045045400000001
+   | OCTET_STRING     bytes
+   | CONTEXT (7)      bytes  CONSTRUCTED
+   | | INTEGER          int    1058588941
+   | | INTEGER          int    0
+   | | INTEGER          int    0
+   | | SEQUENCE
+   | | | SEQUENCE
+   | | | | OID              oid    1.3.6.1.2.1.1.3.0
+   | | | | TIMETICKS        int    638085796
+
+=back
+
+=cut
+
+# reverse enum, very slow and ugly hack
+sub _re {
+   my ($export_tag, $value) = @_;
+
+   for my $symbol (@{ $EXPORT_TAGS{$export_tag} }) {
+      $value == eval $symbol
+         and return $symbol;
+   }
+
+   "($value)"
+}
+
+$SNMP_PROFILE->set (ASN_APPLICATION, SNMP_COUNTER64 , BER_TYPE_INT);
+
+sub _ber_dump {
+   my ($ber, $profile, $indent) = @_;
+
+   if (my $seq = ber_is_seq $ber) {
+      printf "%sSEQUENCE\n", $indent;
+      &_ber_dump ($_, $profile, "$indent| ")
+         for @$seq;
+   } else {
+      my $asn = $ber->[BER_CLASS] == ASN_UNIVERSAL;
+
+      my $class = _re const_asn_class => $ber->[BER_CLASS];
+      my $tag   = $asn ? _re const_asn_tag => $ber->[BER_TAG] : $ber->[BER_TAG];
+      my $type  = _re const_ber_type => $profile->get ($ber->[BER_CLASS], $ber->[BER_TAG]);
+      my $data  = $ber->[BER_DATA];
+
+      if ($profile == $SNMP_PROFILE and $ber->[BER_CLASS] == ASN_APPLICATION) {
+         $tag = _re const_snmp => $ber->[BER_TAG];
+      } elsif (!$asn) {
+         $tag = "$class ($tag)";
+      }
+
+      $class =~ s/^ASN_//;
+      $tag   =~ s/^(ASN_|SNMP_)//;
+      $type  =~ s/^BER_TYPE_//;
+
+      if ($ber->[BER_FLAGS]) {
+         printf "$indent%-16.16s %-6.6s CONSTRUCTED\n", $tag, lc $type;
+         &_ber_dump ($_, $profile, "$indent| ")
+            for @$data;
+      } else {
+         if ($data =~ y/\x20-\x7e//c / (length $data || 1) > 0.2 or $data =~ /\x00./s) {
+            # assume binary
+            $data = unpack "H*", $data;
+         } else {
+            $data =~ s/[^\x20-\x7e]/./g;
+            $data = "\"$data\"" if $type =~ /string/i || !length $data;
+         }
+
+         substr $data, 40, 1e9, "..." if 40 < length $data;
+
+         printf "$indent%-16.16s %-6.6s %s\n", $tag, lc $type, $data;
+      }
+   }
+}
+
+sub ber_dump($;$$) {
+   _ber_dump $_[0], $_[1] || $DEFAULT_PROFILE, $_[2];
+}
 
 =head1 PROFILES
 
@@ -521,10 +643,12 @@ Returns the BER type mapped to the given C<$class>/C<$tag> combination.
 
 =back
 
-=head2 BER TYPES
+=head2 BER Types
 
-This lists the predefined BER types - you can map any C<CLASS>/C<TAG>
-combination to any C<BER_TYPE_*>.
+This lists the predefined BER types. BER types are formatters used
+internally to format and encode BER values. You can assign any C<BER_TYPE>
+to any C<CLASS>/C<TAG> combination tgo change how that tag is decoded or
+encoded.
 
 =over
 
@@ -614,35 +738,18 @@ to the C<$Convert::BER::XS::SNMP_PROFILE> profile.
    $SNMP_PROFILE->set (ASN_APPLICATION, SNMP_OPAQUE    , BER_TYPE_IPADDRESS);
    $SNMP_PROFILE->set (ASN_APPLICATION, SNMP_COUNTER64 , BER_TYPE_INT);
 
-=cut
-
-our $DEFAULT_PROFILE = new Convert::BER::XS::Profile;
-
-$DEFAULT_PROFILE->_set_default;
-
-# additional SNMP application types
-our $SNMP_PROFILE = new Convert::BER::XS::Profile;
-
-$SNMP_PROFILE->set (ASN_APPLICATION, SNMP_IPADDRESS , BER_TYPE_IPADDRESS);
-$SNMP_PROFILE->set (ASN_APPLICATION, SNMP_COUNTER32 , BER_TYPE_INT);
-$SNMP_PROFILE->set (ASN_APPLICATION, SNMP_UNSIGNED32, BER_TYPE_INT);
-$SNMP_PROFILE->set (ASN_APPLICATION, SNMP_TIMETICKS , BER_TYPE_INT);
-$SNMP_PROFILE->set (ASN_APPLICATION, SNMP_OPAQUE    , BER_TYPE_IPADDRESS);
-$SNMP_PROFILE->set (ASN_APPLICATION, SNMP_COUNTER64 , BER_TYPE_INT);
-
-1;
-
 =head2 LIMITATIONS/NOTES
 
 This module can only en-/decode 64 bit signed and unsigned integers, and
-only when your perl supports those.
+only when your perl supports those. So no UUID OIDs for now (unless you
+map the C<OBJECT IDENTIFIER> tag to something other than C<BER_TYPE_OID>).
 
 This module does not generally care about ranges, i.e. it will happily
 de-/encode 64 bit integers into an C<ASN_INTEGER> value, or a negative
 number into an C<SNMP_COUNTER64>.
 
 OBJECT IDENTIFIEERs cannot have unlimited length, although the limit is
-much larger than e.g. the one imposed by SNMP or other protocols,a nd is
+much larger than e.g. the one imposed by SNMP or other protocols, and is
 about 4kB.
 
 Indefinite length encoding is not supported.
@@ -653,7 +760,8 @@ join them for convenience.
 REAL values are not supported and will currently croak.
 
 The encoder and decoder tend to accept more formats than should be
-strictly supported.
+strictly supported - security sensitive applications are strongly advised
+to review the code first.
 
 This module has undergone little to no testing so far.
 
@@ -669,3 +777,5 @@ are in use.
 
 =cut
 
+1;
+
